@@ -67,27 +67,32 @@
       ;; Get and insert candidate
       (let* ((candidates (citar-get-entries))
              (bounds (bounds-of-thing-at-point 'word)))
-    (when bounds
-      (list (car bounds)
-            (cdr bounds)
-            candidates
-            :annotation-function #'citar-capf-annotate
-            :exit-function
-            (lambda (_str _status)
-              (insert))))))))
+        (when bounds
+          (list (car bounds)
+                (cdr bounds)
+                candidates
+                :annotation-function #'citar-capf-annotate
+                :exit-function
+                (lambda (_str _status)
+                  (insert))))))))
 
-  (defun citar-capf-annotate (citekey)
-    "Annotate a CITEKEY."
-    (let* ((author (citar-get-value "author" citekey))
-           (editor (citar-get-value "editor" citekey))
-           (title (citar-get-value "title" citekey)))
-      (concat
-       "   "
-       (truncate-string-to-width
-        (citar--shorten-names
-         (or author editor "")) 20 nil 32 t)
-       "  "
-       (truncate-string-to-width (or title "") 40 nil 32))))
+(defun citar-capf-annotate (citekey)
+  "Annotate a CITEKEY."
+  (let* ((author (citar-get-value "author" citekey))
+         (editor (citar-get-value "editor" citekey))
+         (title (citar-get-value "title" citekey)))
+    (concat
+     "   "
+     (truncate-string-to-width
+      (citar--shorten-names
+       (or author editor "")) 20 nil 32 t)
+     "  "
+     (truncate-string-to-width (or title "") 40 nil 32))))
+
+;;;###autoload
+(defun citar-capf-setup ()
+  "Add `citar-capf' to `completion-at-point-functions'."
+  (add-to-list 'completion-at-point-functions 'citar-capf))
 
 (provide 'citar-capf)
 ;;; citar-capf.el ends here
