@@ -1,12 +1,12 @@
 ;;; citar.el --- Citation-related commands for org, latex, markdown -*- lexical-binding: t; -*-
 
-;; Copyright (C) 2021-2022 Bruce D'Arcus
+;; Copyright (C) 2021-2023 Bruce D'Arcus
 
 ;; Author: Bruce D'Arcus <https://github.com/bdarcus>
 ;; Maintainer: Bruce D'Arcus <https://github.com/bdarcus>
 ;; Created: February 27, 2021
 ;; SPDX-License-Identifier: GPL-3.0-or-later
-;; Version: 1.1
+;; Version: 1.4.0
 ;; Homepage: https://github.com/emacs-citar/citar
 ;; Package-Requires: ((emacs "27.1") (parsebib "4.2") (org "9.5") (citeproc "0.9"))
 
@@ -114,8 +114,8 @@ original entry, but also in entries specified in the field named
 by this variable."
   :group 'citar
   :type '(choice (const "crossref")
-                 (string :tag "Field name")
-                 (const :tag "Ignore cross-references" nil)))
+          (string :tag "Field name")
+          (const :tag "Ignore cross-references" nil)))
 
 (defcustom citar-additional-fields nil
   "A list of fields to add to parsed data.
@@ -130,19 +130,19 @@ specifies additional fields to include."
 ;;;; Displaying completions and formatting
 
 (defcustom citar-templates
-  '((main . "${author editor:30}     ${date year issued:4}     ${title:48}")
+  '((main . "${author editor:30%sn}     ${date year issued:4}     ${title:48}")
     (suffix . "          ${=key= id:15}    ${=type=:12}    ${tags keywords keywords:*}")
-    (preview . "${author editor} (${year issued date}) ${title}, \
+    (preview . "${author editor:%etal} (${year issued date}) ${title}, \
 ${journal journaltitle publisher container-title collection-title}.\n")
-    (note . "Notes on ${author editor}, ${title}"))
+    (note . "Notes on ${author editor:%etal}, ${title}"))
   "Configures formatting for the bibliographic entry.
 
 The main and suffix templates are for candidate display, and note
 for the title field for new notes."
   :group 'citar
   :type  '(alist :key-type symbol
-                 :value-type string
-                 :options (main suffix preview note)))
+           :value-type string
+           :options (main suffix preview note)))
 
 (defcustom citar-ellipsis nil
   "Ellipsis string to mark ending of truncated display fields.
@@ -152,10 +152,24 @@ ellipsis will be used.  Otherwise, this should be a non-empty
 string specifying the ellipsis."
   :group 'citar
   :type '(choice (const :tag "Use `truncate-string-ellipsis'" t)
-                 (const :tag "No ellipsis" nil)
-                 (const "…")
-                 (const "...")
-                 (string :tag "Ellipsis string")))
+          (const :tag "No ellipsis" nil)
+          (const "…")
+          (const "...")
+          (string :tag "Ellipsis string")))
+
+(defcustom citar-open-entry-function #'citar-open-entry-in-file
+  "The function to run for `citar-open-entry'.
+
+This function must accept one argument; the citekey."
+  :group 'citar
+  :type '(choice
+          (function-item
+           :tag "Open in BibTeX/BibLaTeX/CSL JSON file"
+           citar-open-entry-in-file)
+          (function-item
+           :tag "Open in Zotero/Better BibTeX"
+           citar-open-entry-in-zotero)
+          (function :tag "Other")))
 
 (defcustom citar-format-reference-function
   #'citar-format-reference
@@ -171,23 +185,100 @@ function that takes a list of CITEKEYS and returns formatted
 references as a string."
   :group 'citar
   :type '(choice (function-item :tag "Use 'citar-template'" citar-format-reference)
-                 (function-item :tag "Use 'citeproc-el'" citar-citeproc-format-reference)
-                 (function :tag "Other")))
+          (function-item :tag "Use 'citeproc-el'" citar-citeproc-format-reference)
+          (function :tag "Other")))
 
 (defcustom citar-display-transform-functions
   ;; TODO change this name, as it might be confusing?
-  '((("author" "editor") . citar--shorten-names))
+  `((sn . (citar--shorten-names))
+    (etal . (citar--shorten-names 3 "&")))
   "Configure transformation of field display values from raw values.
 
-All functions that match a particular field are run in order."
+When the car symbol is associated with a field, run the cdr function
+and optional arguments on the string value."
   :group 'citar
-  :type '(alist :key-type   (choice (const t) (repeat string))
-                :value-type function))
+  :type '(alist :key-type symbol
+          ;; REVIEW is this OK for now?
+          :value-type list))
+
+;; Indicator defstruct
+
+;;;###autoload
+(cl-defstruct
+ (citar-indicator (:constructor citar-indicator-create)
+                  (:copier nil))
+ "A citar indicator specification."
+ (tag
+  nil
+  :documentation
+  "The string to include as hidden candidate text, and to then determine whether a
+candidate predicate function will return non-nil.")
+ (symbol
+  nil
+  :type string
+  :documentation
+  "The symbol string to use in the UI when predicate function returns non-nil.")
+ (padding
+  " "
+  :type string
+  :documentation
+  "String to add to the right side of the indicator, for proper padding and such.")
+ (emptysymbol
+  ;; REVIEW we may not need this, so perhaps remove?
+  " "
+  :documentation
+  "The symbol to use in the UI when predicate function returns nil. Can be useful
+in some cases when using icons.")
+ (function
+  nil
+  :type function
+  :documentation
+  "A predicate function that takes a single CITEKEY argument.")
+ (compiledfunction
+  nil
+  :type compiled-function
+  :documentation
+  "A compiled version of `function' used during processing."))
+
+;; Indicator specs
+
+(defvar citar-indicator-files
+   (citar-indicator-create
+    :symbol "F"
+    :function #'citar-has-files
+    :tag "has:files"))
+
+(defvar citar-indicator-links
+   (citar-indicator-create
+    :symbol "L"
+    :function #'citar-has-links
+    :tag "has:links"))
+
+(defvar citar-indicator-notes
+  (citar-indicator-create
+   :symbol "N"
+   :function #'citar-has-notes
+   :tag "has:notes"))
+
+(defvar citar-indicator-cited
+  (citar-indicator-create
+   :symbol "C"
+   :function #'citar-is-cited
+   :tag "is:cited"))
+
+;; Indicator config
+
+(defvar citar-indicators
+  (list citar-indicator-links
+        citar-indicator-files
+        citar-indicator-notes
+        citar-indicator-cited))
 
 (defcustom citar-symbols
   `((file  .  ("F" . " "))
     (note .   ("N" . " "))
     (link .   ("L" . " ")))
+  ;; DEPRECATED
   "Configuration alist specifying which symbol or icon to pick for a bib entry.
 This leaves room for configurations where the absense of an item
 may be indicated with the same icon but a different face.
@@ -196,15 +287,19 @@ To avoid alignment issues make sure that both the car and cdr of a symbol have
 the same width."
   :group 'citar
   :type '(alist :key-type symbol
-                :value-type (cons (string :tag "Present")
-                                  (string :tag "Absent"))
-                :options (file note link)))
+          :value-type (cons (string :tag "Present")
+                            (string :tag "Absent"))
+          :options (file note link)))
 
 (defcustom citar-symbol-separator " "
   "The padding between prefix symbols."
+  ;; DEPRECATED
   :group 'citar
   :type 'string)
 
+(make-obsolete 'citar-symbols nil "1.4")
+(make-obsolete 'citar-symbol-separator nil "1.4")
+
 ;;;; Citar actions and other miscellany
 
 (defcustom citar-default-action #'citar-open
@@ -221,7 +316,7 @@ The action is used when no citation key is found at point.
 and nil means no action."
   :group 'citar
   :type '(radio (const :tag "Prompt" prompt)
-                (const :tag "Ignore" nil)))
+          (const :tag "Ignore" nil)))
 
 (defcustom citar-open-prompt (list #'citar-open #'citar-attach-files #'citar-open-note)
   "Always prompt to open files, notes, or links.
@@ -236,14 +331,14 @@ Otherwise, the value should be a list of command names that will
 always prompt to select."
   :group 'citar
   :type '(choice (const :tag "Always prompt" t)
-                 (const :tag "Prompt only for multiple resources" nil)
-                 (set :tag "Commands that prompt for multiple resources"
-                      (function-item citar-open)
-                      (function-item citar-open-files)
-                      (function-item citar-attach-files)
-                      (function-item citar-open-links)
-                      (function-item citar-open-notes)
-                      (function-item citar-open-note))))
+          (const :tag "Prompt only for multiple resources" nil)
+          (set :tag "Commands that prompt for multiple resources"
+               (function-item citar-open)
+               (function-item citar-open-files)
+               (function-item citar-attach-files)
+               (function-item citar-open-links)
+               (function-item citar-open-notes)
+               (function-item citar-open-note))))
 
 ;;;; File, note, and URL handling
 
@@ -251,9 +346,9 @@ always prompt to select."
   "Types of resources that `citar-open' offers to open."
   :group 'citar
   :type '(set (const :tag "Library files" :files)
-              (const :tag "Links" :links)
-              (const :tag "Notes" :notes)
-              (const :tag "Create notes" :create-notes)))
+          (const :tag "Links" :links)
+          (const :tag "Notes" :notes)
+          (const :tag "Create notes" :create-notes)))
 
 (defcustom citar-open-always-create-notes nil
   "Offer to create notes even for keys that already have notes.
@@ -266,15 +361,16 @@ Otherwise, the value should be a list of command names that will
 offer to create new notes unconditionally."
   :group 'citar
   :type '(choice (const :tag "Always offer to create notes" t)
-                 (const :tag "Create notes only if none exist" nil)
-                 (set :tag "Create notes for commands"
-                      (function-item citar-open)
-                      (function-item citar-open-notes))))
+          (const :tag "Create notes only if none exist" nil)
+          (set :tag "Create notes for commands"
+               (function-item citar-open)
+               (function-item citar-open-notes))))
 
-(defcustom citar-file-sources (list (list :items #'citar-file--get-from-file-field
-                                          :hasitems #'citar-file--has-file-field)
-                                    (list :items #'citar-file--get-library-files
-                                          :hasitems #'citar-file--has-library-files))
+(defcustom citar-file-sources
+  (list (list :items #'citar-file--get-from-file-field
+              :hasitems #'citar-file--has-file-field)
+        (list :items #'citar-file--get-library-files
+              :hasitems #'citar-file--has-library-files))
   "List of backends used to get library files for bibliography references.
 
 Should be a list of plists, where each plist has the following properties:
@@ -287,15 +383,67 @@ Should be a list of plists, where each plist has the following properties:
   :group 'citar
   :type '(repeat (plist :value-type function :options (:items :hasitems))))
 
+(defcustom citar-add-file-sources
+  '((?b "buffer" "Current buffer" citar--add-file-from-buffer)
+    (?f "file" "Existing file" citar--add-file-from-file)
+    (?u "url" "Download from URL" citar--add-file-from-url))
+  "List of sources from which library files can be added.
+
+The command `citar-add-file-to-library' prompts for the source of
+the file to add, with the options being the elements of this
+list. Each source must be a list containing:
+
+  - The unique shortcut character the user types to select the
+    source.
+
+  - Short name for the source, displayed in the prompt.
+
+  - Optional longer description, displayed if the user requests
+    help.
+
+  - Function that provides the file to add, described next.
+
+The function is called with a bibliography key and should return
+a plist containing the following keys:
+
+  :extension The extension of the source file if known, or nil.
+
+  :write-file A function taking arguments DESTFILE and
+  OK-IF-ALREADY-EXISTS. When called, it should write the source
+  file into DESTFILE. If DESTFILE already exists, it should act
+  according to OK-IF-ALREADY-EXISTS, emulating `copy-file'.
+
+The plist returned by this function is passed to
+`citar-add-file-function', which see."
+  :group 'citar
+  :type '(repeat :tag "Sources for `citar-add-file-to-library'"
+                 (group (character :tag "Shortcut")
+                        (string :tag "Name")
+                        (string :tag "Description")
+                        (function :tag "Source function"))))
+
+(defcustom citar-add-file-function #'citar-save-file-to-library
+  "Function run by `citar-add-file-to-library' to add a file.
+
+This function must accept two arguments: the CITEKEY of a
+bibliography item; and the SOURCE-PLIST documented in
+`citar-add-file-sources'."
+  :group 'citar
+  :type '(radio
+          (function-item
+           :doc "Save the file to `citar-library-paths'."
+           citar-save-file-to-library)
+          (function :tag "Other")))
+
 (defcustom citar-notes-sources
   `((citar-file .
-                ,(list :name "Notes"
-                       :category 'file
-                       :items #'citar-file--get-notes
-                       :hasitems #'citar-file--has-notes
-                       :open #'find-file
-                       :create #'citar-file--create-note
-                       :transform #'file-name-nondirectory)))
+     ,(list :name "Notes"
+            :category 'file
+            :items #'citar-file--get-notes
+            :hasitems #'citar-file--has-notes
+            :open #'find-file
+            :create #'citar-file--create-note
+            :transform #'file-name-nondirectory)))
   "The alist of notes backends available for configuration.
 
 The format of the cons should be (NAME . PLIST), where the
@@ -651,17 +799,17 @@ CATEGORY is one of:
   (cl-flet ((getresources (table) (when table
                                     (delete-dups (apply #'append (hash-table-values table)))))
             (keycands (type citekeys)
-                      (let ((format (citar-format--parse (citar--get-template 'completion)))
-                            (width (- (frame-width) 2)))
-                        (mapcar (lambda (key)
-                                  (let* ((entry (citar-get-entry key))
-                                         (cand (citar-format--entry format entry width
-                                                                    :ellipsis citar-ellipsis))
-                                         (keycand (citar--prepend-candidate-citekey key cand))
-                                         (target (cons 'citar-reference
-                                                       (propertize key 'citar--resource type))))
-                                    (propertize keycand 'multi-category target)))
-                                citekeys)))
+              (let ((format (citar-format--parse (citar--get-template 'completion)))
+                    (width (- (frame-width) 2)))
+                (mapcar (lambda (key)
+                          (let* ((entry (citar-get-entry key))
+                                 (cand (citar-format--entry format entry width
+                                                            :ellipsis citar-ellipsis))
+                                 (keycand (citar--prepend-candidate-citekey key cand))
+                                 (target (cons 'citar-reference
+                                               (propertize key 'citar--resource type))))
+                            (propertize keycand 'multi-category target)))
+                        citekeys)))
             (withtype (type cat cands) (when cands
                                          (cons cat (mapcar (lambda (cand)
                                                              (propertize cand 'citar--resource type))
@@ -752,6 +900,62 @@ only one resource and `citar-open-prompt' is t or contains
            resource
          nil))))
 
+;; Indicator functions
+
+(defun citar--make-indicator-processors (ispecs)
+  "Set the compiledfunction slots in ISPECS."
+  (mapc
+   (lambda (ispec)
+     (let ((fnsym (citar-indicator-function ispec)))
+       (setf (citar-indicator-compiledfunction ispec)
+             (funcall fnsym))))
+   ispecs))
+
+(defun citar--make-indicator-tags (citekey iprocs)
+  "Return indicator tags string for CITEKEY, using IPROCS.
+
+This string is incorporated in the candidates as hidden text, so
+it can be searched against, and to contruct the indicator symbols
+visible in the completion UI."
+  (mapconcat
+   (lambda (iproc)
+     (when-let* ((cfun (citar-indicator-compiledfunction iproc))
+                 (tag (citar-indicator-tag iproc)))
+       (when (funcall cfun citekey)
+         (propertize (concat " " tag) 'invisible t))))
+   iprocs ""))
+
+(defun citar--make-indicator-symbols (candidate)
+  "Return indicator string for CANDIDATE display."
+  (seq-reduce
+    (lambda (constructed ispec)
+      ;; first arg is the accumulated string
+      (let* ((matchtext (citar-indicator-tag ispec))
+             (matchtagp (string-match-p matchtext candidate))
+             (sym (citar-indicator-symbol ispec))
+             (emptysym (citar-indicator-emptysymbol ispec))
+             (padding (citar-indicator-padding ispec))
+             (str (concat
+                   constructed
+                   (if matchtagp sym emptysym)
+                   padding))
+             (pos (length str)))
+        ;; See https://github.com/emacs-citar/citar/issues/764 for explanation of what this code is doing.
+        ;;
+        ;; We say that the last character of the padding should be replaced by a space which stretches to the
+        ;; position we want to.
+        ;;
+        ;; So for example if the icon has string width 1 but occupies 0.9 the space will stretch to occupy 1.1
+        ;; and if the icon occupies 1.5 it will shrink to occupy 0.5.
+        ;;
+        ;; Emacs 29 has `vtable', which we might use here in the future as well.
+        (put-text-property (- pos 1) pos 'display
+                           (cons 'space
+                                 (list :align-to (string-width str)))
+                           str)
+        str))
+    citar-indicators ""))
+
 (defun citar--format-candidates ()
   "Format completion candidates for bibliography entries.
 
@@ -763,32 +967,21 @@ Return nil if `citar-bibliographies' returns nil."
   (when-let ((bibs (citar--bibliographies)))
     (let* ((citar--entries (citar-cache--entries bibs))
            (preformatted (citar-cache--preformatted bibs))
-           (hasfilesp (citar-has-files))
-           (hasnotesp (citar-has-notes))
-           (haslinksp (citar-has-links))
-           (hasfilestag (propertize " has:files" 'invisible t))
-           (hasnotestag (propertize " has:notes" 'invisible t))
-           (haslinkstag (propertize " has:links" 'invisible t))
-           (symbolswidth (string-width (citar--symbols-string t t t)))
+           (indicatorprocs (citar--make-indicator-processors citar-indicators))
+           (symbolswidth (string-width
+                          (citar--make-indicator-symbols "")))
            (width (- (frame-width) symbolswidth 2))
            (completions (make-hash-table :test 'equal :size (hash-table-count citar--entries))))
       (prog1 completions
         (maphash
          (lambda (citekey _entry)
-           (let* ((hasfiles (and hasfilesp (funcall hasfilesp citekey)))
-                  (hasnotes (and hasnotesp (funcall hasnotesp citekey)))
-                  (haslinks (and haslinksp (funcall haslinksp citekey)))
-                  (preform (or (gethash citekey preformatted)
+           (let* ((preform (or (gethash citekey preformatted)
                                (error "No preformatted candidate string: %s" citekey)))
                   (display (citar-format--star-widths
                             (- width (car preform)) (cdr preform)
                             t citar-ellipsis))
-                  (tagged (if (not (or hasfiles hasnotes haslinks))
-                              display
-                            (concat display
-                                    (when hasfiles hasfilestag)
-                                    (when hasnotes hasnotestag)
-                                    (when haslinks haslinkstag)))))
+                  (tagged (concat display
+                                  (citar--make-indicator-tags citekey indicatorprocs))))
              (puthash tagged citekey completions)))
          citar--entries)))))
 
@@ -900,19 +1093,19 @@ was found to have a value, and VALUE is its value."
                   (cons field value)))
               fields)))
 
-(defun citar-get-display-value (fields citekey-or-entry)
+(defun citar-get-display-value (fields citekey-or-entry &optional transform)
   "Return the first non nil value for CITEKEY-OR-ENTRY among FIELDS .
 
-The value is transformed using `citar-display-transform-functions'"
-  (let ((fieldvalue (citar-get-field-with-value fields citekey-or-entry)))
-    (seq-reduce (lambda (string fun)
-                  (if (or (eq t (car fun))
-                          (seq-contains-p (car fun) (car fieldvalue) #'string=))
-                      (funcall (cdr fun) string)
-                    string))
-                citar-display-transform-functions
-                ;; Make sure we always return a string, even if empty.
-                (or (cdr fieldvalue) ""))))
+When TRANSFORM, use the list of function symbol and optional
+arguments to transform the string."
+  (let* ((fieldvalue (citar-get-field-with-value fields citekey-or-entry))
+         (fnsymbol (car transform))
+         (fnargs (cdr transform))
+         (tvalue (if transform
+                     (apply fnsymbol (cdr fieldvalue) fnargs)
+                   (cdr fieldvalue))))
+    ;; Make sure we always return a string, even if empty.
+    (or tvalue "")))
 
 ;;;; File, notes, and links
 
@@ -990,6 +1183,13 @@ nil, return nil."
                                      (push (format urlformat fieldvalue) keylinks))))
                                (nreverse keylinks)))))))
 
+(defun citar-is-cited ()
+  "Return function to check if reference is cited in buffer."
+  (let ((iscited
+         (citar--major-mode-function 'list-keys #'ignore)))
+    (lambda (citekey)
+      (member citekey iscited))))
+
 (defun citar-has-files ()
   "Return predicate testing whether entry has associated files.
 
@@ -1134,20 +1334,125 @@ getters for file, note, and link resources."
           (dolist (citekey citekeys)
             (putresult citekey (citar-get-entry citekey))))))))
 
+(defun citar-save-file-to-library (citekey source-plist)
+  "Use as `citar-add-file-function' to save files to `citar-library-paths'.
+
+Save the file with the filename CITEKEY.EXTENSION to a directory
+from `citar-library-paths'. If there are multiple such
+directories, prompt the user to choose one. If SOURCE-PLIST has
+an :extension key, use its value as EXTENSION; otherwise prompt
+the user. If the file already exists, confirm before overwriting
+it.
+
+SOURCE-PLIST must be as specified in the documentation of
+`citar-add-file-sources'."
+  (let* ((directory (if (cdr citar-library-paths)
+                        (completing-read "Directory: " citar-library-paths)
+                      (car citar-library-paths)))
+         (extension (or (plist-get source-plist :extension)
+                        (read-string "File extension: ")))
+         (destfile (expand-file-name citekey directory))
+         (destfile (if (string-empty-p extension)
+                       destfile
+                     (concat destfile "." extension)))
+         ;; an integer means to confirm before overwriting
+         (ok-if-already-exists 1))
+    (funcall (plist-get source-plist :write-file)
+             destfile ok-if-already-exists)))
+
+(defun citar--add-file-from-buffer (_citekey)
+  "Include in `citar-add-file-sources' to add contents of current buffer.
+
+See the documentation for `citar-add-file-sources' for more details."
+  (let ((buf (get-buffer (read-buffer "Add file buffer: " (current-buffer)))))
+    (list :write-file
+          (lambda (destfile ok-if-already-exists)
+            (with-current-buffer buf
+              (if (and buffer-file-name
+                       (file-equal-p destfile buffer-file-name))
+                  (if (not (buffer-modified-p))
+                      (message "%s exists and the current buffer is visiting it."
+                               (file-name-nondirectory buffer-file-name))
+                    (unless ok-if-already-exists
+                      (signal 'file-already-exists
+                              (list "File already exists" destfile)))
+                    (when (or (not (integerp ok-if-already-exists))
+                              (yes-or-no-p
+                               (format
+                                "%s exists and the current buffer is visiting it.  Save anyway? "
+                                (file-name-nondirectory buffer-file-name))))
+                      (save-buffer)))
+                (write-region nil nil destfile
+                              (if ok-if-already-exists
+                                  ;; Confirm if integer, otherwise overwrite silently:
+                                  (integerp ok-if-already-exists)
+                                ;; Otherwise signal 'file-already-exists error
+                                'excl)))))
+          :extension
+          (when (buffer-file-name buf)
+            (file-name-extension (buffer-file-name buf))))))
+
+(defun citar--add-file-from-file (_citekey)
+  "Include in `citar-add-file-sources' to add copy of existing file.
+
+See the documentation for `citar-add-file-sources' for more details."
+  (let ((file (read-file-name "Add file: " nil nil t)))
+    (list :write-file
+          (lambda (destfile ok-if-already-exists)
+            (copy-file file destfile ok-if-already-exists))
+          :extension (file-name-extension file))))
+
+(defun citar--add-file-from-url (_citekey)
+  "Include in `citar-add-file-sources' to add from URL.
+
+See the documentation for `citar-add-file-sources' for more details."
+  (let* ((url (read-string "Add file URL: "))
+         ;; TODO: Use Content-Type HTTP response header to guess file extension
+         (dot-ext (url-file-extension url)))
+    (list :write-file
+          (lambda (destfile ok-if-already-exists)
+            (url-copy-file url destfile ok-if-already-exists))
+          :extension (when (> (length dot-ext) 1) (substring dot-ext 1)))))
+
 ;;; Format and display field values
 
-(defun citar--shorten-names (names)
-  "Return a list of family names from a list of full NAMES.
+(defun citar--shorten-name-position (namelist name)
+  "Return NAME position in a NAMELIST."
+  (+ (seq-position namelist name) 1))
 
+(defun citar--shorten-name (name)
+  "Return family NAME in `family, given' string.
+
+Otherwise, return as is."
+  (car (split-string name ", ")))
+
+(defun citar--shorten-names (namestr &optional truncate andstr)
+  "Return a list of family names from a list of full NAMESTR.
 To better accommodate corporate names, this will only shorten
-personal names of the form \"family, given\"."
-  (when (stringp names)
+personal names of the form \"family, given\".
+
+With an integer TRUNCATE, will shorten the list, and ANDSTR will
+replace last comma."
+  (let* ((namelist (split-string (or namestr "") " and "))
+         (namelength (length namelist))
+         (tnamelist (seq-take namelist (or truncate namelength)))
+         (tnamelength (length tnamelist)))
     (mapconcat
-     (lambda (name)
-       (if (eq 1 (length name))
-           (cdr (split-string name " "))
-         (car (split-string name ", "))))
-     (split-string names " and ") ", ")))
+     (lambda (n)
+       (let* ((shortname (citar--shorten-name n))
+              (pos (citar--shorten-name-position tnamelist n))
+              (suffix
+               (cond
+                ;; if last name in the list and we're truncating add et al.; otherwise, no suffix
+                ((equal pos tnamelength)
+                 (if (< tnamelength namelength) " et al." ""))
+                ;; if second to last in the list, and ANDSTR, use that
+                ((and andstr (equal pos (- tnamelength 1)))
+                 (concat " " andstr " "))
+                ;; otherwise, use a comma
+                (t ", "))))
+         (concat shortname suffix)))
+     tnamelist "")))
 
 (defun citar--fields-for-format (template)
   "Return list of fields for TEMPLATE."
@@ -1174,46 +1479,19 @@ personal names of the form \"family, given\"."
 
 ;;; Affixations and annotations
 
-(defun citar--ref-affix (cands)
-  "Add affixation prefix to CANDS."
+(defun citar--ref-affix (candidates)
+  "Add affixation prefix to CANDIDATES."
   (seq-map
    (lambda (candidate)
-     (let ((symbols (citar--ref-make-symbols candidate)))
+     (let ((symbols
+            (citar--make-indicator-symbols candidate)))
        (list candidate symbols "")))
-   cands))
+   candidates))
 
-(defun citar--ref-make-symbols (cand)
-  "Make CAND annotation or affixation string for has-symbols."
-  (let ((candidate-symbols (citar--symbols-string
-                            (string-match-p "has:files" cand)
-                            (string-match-p "has:notes" cand)
-                            (string-match-p "has:links" cand))))
-    candidate-symbols))
-
-(defun citar--ref-annotate (cand)
-  "Add annotation to CAND."
+(defun citar--ref-annotate (candidate)
+  "Add annotation to CANDIDATE."
   ;; REVIEW/TODO we don't currently use this, but could, for Emacs 27.
-  (citar--ref-make-symbols cand))
-
-(defun citar--symbols-string (has-files has-note has-link)
-  "String for display from booleans HAS-FILES HAS-LINK HAS-NOTE."
-  (cl-flet ((thing-string (has-thing thing-symbol)
-                          (if has-thing
-                              (cadr (assoc thing-symbol citar-symbols))
-                            (cddr (assoc thing-symbol citar-symbols)))))
-    (seq-reduce (lambda (constructed newpart)
-                  (let* ((str (concat constructed newpart
-                                      citar-symbol-separator))
-                         (pos (length str)))
-                    (put-text-property (- pos 1) pos 'display
-                                       (cons 'space (list :align-to (string-width str)))
-                                       str)
-                    str))
-                (list (thing-string has-files 'file)
-                      (thing-string has-note 'note)
-                      (thing-string has-link 'link)
-                      "")
-                "")))
+  (citar--make-indicator-symbols candidate))
 
 (defun citar--get-template (template-name)
   "Return template string for TEMPLATE-NAME."
@@ -1345,26 +1623,34 @@ to open from a list of all notes."
 (defun citar-open-entry (citekey)
   "Open bibliographic entry associated with the CITEKEY."
   (interactive (list (citar-select-ref)))
-  (when-let ((bib-files (citar--bibliography-files)))
-    (citar--open-entry citekey bib-files)))
+  (funcall citar-open-entry-function citekey))
 
-(defun citar--open-entry (citekey bib-files)
-  "Open entry for CITEKEY in the relevant BIB-FILES."
+(defun citar-open-entry-in-file (citekey)
+  "Open entry for CITEKEY."
   ;; Adapted from 'bibtex-completion-show-entry'.
-  (catch 'break
-    (dolist (bib-file bib-files)
-      (let ((buf (or (get-file-buffer bib-file)
-                     (find-buffer-visiting bib-file))))
-        (find-file bib-file)
-        (widen)
-        (goto-char (point-min))
-        (when (re-search-forward
-               (concat "^@\\(" parsebib--bibtex-identifier
-                       "\\)[[:space:]]*[\\(\\{][[:space:]]*"
-                       (regexp-quote citekey) "[[:space:]]*,") nil t)
-          (throw 'break t))
-        (unless buf
-          (kill-buffer))))))
+  (when-let ((bib-files (citar--bibliography-files)))
+    (catch 'break
+      (dolist (bib-file bib-files)
+        (let ((buf (or (get-file-buffer bib-file)
+                       (find-buffer-visiting bib-file))))
+          (find-file bib-file)
+          (widen)
+          (goto-char (point-min))
+          (when (re-search-forward
+                 (concat "^@\\(" parsebib--bibtex-identifier
+                         "\\)[[:space:]]*[\\(\\{][[:space:]]*"
+                         (regexp-quote citekey) "[[:space:]]*,") nil t)
+            (throw 'break t))
+          (unless buf
+            (kill-buffer)))))))
+
+(defun citar-open-entry-in-zotero (citekey)
+  "Open a reference item for CITEKEY in Zotero.
+
+This function assumes a setup where the bibliographic data,
+including the citekeys, is maintained in Zotero with Better BibTeX."
+  (citar-file-open-external
+   (concat "zotero://select/items/@" citekey)))
 
 ;;;###autoload
 (defun citar-insert-bibtex (citekeys)
@@ -1393,7 +1679,7 @@ to open from a list of all notes."
 (defun citar-export-local-bib-file ()
   "Create a new bibliography file from citations in current buffer.
 
-The file is titled \"local-bib\", given the same extention as
+The file is titled \"local-bib\", given the same extension as
 the first entry in `citar-bibliography', and created in the same
 directory as current buffer."
   (interactive)
@@ -1481,33 +1767,15 @@ ARG is forwarded to the mode-specific insertion function given in
 The FILE can be added from an open buffer, a file path, or a
 URL."
   (interactive (list (citar-select-ref)))
-  (citar--check-configuration 'citar-library-paths)
+  (citar--check-configuration 'citar-library-paths 'citar-add-file-sources
+                              'citar-add-file-function)
   (unless citar-library-paths
     (user-error "Make sure `citar-library-paths' is non-nil"))
-  (let* ((directory (if (cdr citar-library-paths)
-                        (completing-read "Directory: " citar-library-paths)
-                      (car citar-library-paths)))
-         (filepath (expand-file-name citekey directory))
-         (withext (lambda (extension)
-                    (let* ((extension (or extension (read-string "File extension: "))))
-                      (if (string-empty-p extension)
-                          filepath
-                        (concat filepath "." extension))))))
-    (pcase (read-char-choice "Add file from [b]uffer, [f]ile, or [u]rl? " '(?b ?f ?u))
-      (?b
-       (with-current-buffer (read-buffer "Add file buffer: " (current-buffer))
-         (let ((destfile (funcall withext (and buffer-file-name (file-name-extension buffer-file-name)))))
-           (write-file destfile 'confirm))))
-      (?f
-       (let* ((file (read-file-name "Add file: " nil nil t))
-              (destfile (funcall withext (file-name-extension file))))
-         (copy-file file destfile 1)))  ; last arg integer means to confirm before overwriting
-      (?u
-       (let* ((url (read-string "Add file URL: "))
-              ;; TODO: Use Content-Type HTTP response header to guess file extension
-              (ext (url-file-extension url))
-              (destfile (funcall withext (when (> (length ext) 1) (substring ext 1)))))
-         (url-copy-file url destfile 1))))))
+  (unless citar-add-file-sources
+    (user-error "Make sure `citar-add-file-sources' is non-nil"))
+  (let ((source (read-multiple-choice "Add file from" citar-add-file-sources)))
+    (funcall citar-add-file-function
+             citekey (funcall (nth 3 source) citekey))))
 
 ;;;###autoload
 (defun citar-run-default-action (citekeys)
@@ -1541,9 +1809,21 @@ VARIABLES should be the names of Citar customization variables."
         ((or 'citar-has-files-functions 'citar-get-files-functions 'citar-file-parser-functions)
          (unless (and (listp value) (seq-every-p #'functionp value))
            (error "`%s' should be a list of functions: %S" variable `',value)))
-        ((or 'citar-note-format-function)
+        ((or 'citar-note-format-function 'citar-add-file-function)
          (unless (functionp value)
            (error "`%s' should be a function: %S" variable `',value)))
+        ('citar-add-file-sources
+         (unless (listp value)
+           (error "`%s' should be a list: %S" variable `',value))
+         (unless (seq-every-p (lambda (x)
+                                (and (= 4 (length x))
+                                     (characterp (nth 0 x))
+                                     (stringp (nth 1 x))
+                                     (stringp (nth 2 x))
+                                     (functionp (nth 3 x))))
+                              value)
+           (error "Every element of `%s' should be a list of the type (character string string function): %S"
+                  variable `',value)))
         (_
          (error "Unknown variable in citar--check-configuration: %s" variable))))))
 
